#!/usr/bin/env python
"""
Collection of functions to extract spectral features from audio signals
License: New BSD License
"""

# =============================================================================
# Load the modules
# =============================================================================
# Import external modules

import numpy as np
import pandas as pd
from scipy import interpolate
from scipy.optimize import root

# Import internal modules
from maad.util import moments, power2dB
from maad import sound

# define internal functions
def _interpolate_peak_location(pxx):
    """ Estimate peak using quadratic interpolation """
    idxmax = pxx.index.get_loc(pxx.idxmax())
    if (idxmax==0) | (idxmax+1==len(pxx)):
        # if maximum is at beginning or at the end os spectrum, take maximum
        peak = pxx.idxmax()
        amplitude = pxx[peak]
    else:
        # use interpolation
        x = pxx.iloc[[idxmax-1, idxmax, idxmax+1]].index.values
        y = pxx.iloc[[idxmax-1, idxmax, idxmax+1]].values
        f = interpolate.interp1d(x,y, kind='quadratic')
        xnew = np.arange(x[0], x[2], 1)
        ynew = f(xnew)
        peak = xnew[ynew.argmax()]
        amplitude = ynew[ynew.argmax()]
    return peak, amplitude

def _quantile_is_valid(q):
    """ Check if quantile is valid
        function from older version of numpy than 2.0.0
    """
    # avoid expensive reductions, relevant for arrays with < O(1000) elements
    if q.ndim == 1 and q.size < 10:
        for i in range(q.size):
            if not (0.0 <= q[i] <= 1.0):
                return False
    else:
        if not (np.all(0 <= q) and np.all(q <= 1)):
            return False
    return True

#%%
# =============================================================================
# public functions
# =============================================================================
#%%
def spectral_moments (X, axis=None):
    """
    Computes the first 4th moments of an amplitude spectrum (1d) or spectrogram (2d),
    mean, variance, skewness, kurtosis.

    Parameters
    ----------
    X : ndarray of floats
        Amplitude  spectrum (1d) or spectrogram (2d).
    axis : interger, optional, default is None
        if spectrogram (2d), select the axis to estimate the moments.

    Returns
    -------
    mean : float
        mean of the audio
    var : float
        variance  of the audio
    skew : float
        skewness of the audio
    kurt : float
        kurtosis of the audio

    Examples
    --------
    >>> from maad import sound, features
    >>> s, fs = sound.load('../data/spinetail.wav')
    >>> Sxx_power,_,_,_ = sound.spectrogram(s, fs)

    Compute spectral moments on the mean spectrum

    >>> import numpy as np
    >>> S_power = sound.avg_power_spectro(Sxx_power)
    >>> sm, sv, ss, sk = features.spectral_moments (S_power)
    >>> print('mean: %2.8f / var: %2.10f / skewness: %2.2f / kurtosis: %2.2f' % (sm, sv, ss, sk))
    mean: 0.00000228 / var: 0.0000000001 / skewness: 5.84 / kurtosis: 40.49

    Compute spectral moments of the spectrogram along the time axis

    >>> from maad.features import spectral_moments
    >>> sm_per_bin, sv_per_bin, ss_per_bin, sk_per_bin = spectral_moments(Sxx_power, axis=1)
    >>> print('Length of sk_per_bin is : %2.0f' % len(sk_per_bin))
    Length of sk_per_bin is : 512

    """
    # force P to be ndarray
    X = np.asarray(X)

    return moments(X,axis)

#%%
def peak_frequency(s, fs, method='best', nperseg=1024, roi=None, amp=False, as_pandas=False, **kwargs):
    """
    Compute the peak frequency for audio signal. The peak frequency is the frequency of
    maximum power. If a region of interest with time and spectral limits is provided,
    the peak frequency is computed on the selection.

    Parameters
    ----------
    s : 1D array
        Input audio signal
    fs : float
        Sampling frequency of audio signal
    method : {'fast', 'best'}, optional
        Method used to compute the peak frequency.
        The default is 'best'.
    nperseg : int, optional
        Length of segment to compute the FFT. The default is 1024.
    roi : pandas.Series, optional
        Region of interest where peak frequency will be computed.
        Series must have a valid input format with index: min_t, min_f, max_t, max_f.
        The default is None.
    as_pandas: bool
        Return data as a pandas.Series or pandas.DataFrame, when amp
        is False or True, respectively. Default is False.
    kwargs : additional keyword arguments
        Additional keyword arguments to pass to `sound.spectrum`.

    Returns
    -------
    peak_freq : float
        Peak frequency of audio segment.
    amp_peak_freq : float
        Amplitud of the peak frequency of audio segment.

    Examples
    --------
    >>> from maad import features, sound
    >>> s, fs = sound.load('../data/spinetail.wav')

    Compute peak frequency

    >>> peak_freq, peak_freq_amp = features.peak_frequency(s, fs, amp=True)
    >>> print('Peak Frequency: {:.5f}, Amplitude: {:.5f}'.format(peak_freq, peak_freq_amp))
    Peak Frequency: 6634.16016, Amplitude: 0.00013

    """
    if roi is None:
        pxx, fidx = sound.spectrum(s, fs, nperseg, **kwargs)
    else:
        pxx, fidx = sound.spectrum(s, fs, nperseg,
                                   tlims=[roi.min_t, roi.max_t],
                                   flims=[roi.min_f, roi.max_f],
                                   **kwargs)
    pxx = pd.Series(pxx, index=fidx)

    if method=='fast':
        # simplest form to get peak frequency, but less precise
        peak_freq = pxx.idxmax()
        amp_peak_freq = pxx[peak_freq]

    elif method=='best':
        # use interpolation get better estimate of peak frequency
        peak_freq, amp_peak_freq = _interpolate_peak_location(pxx)

    else:
        raise Exception("Invalid method. Method should be 'fast' or 'best' ")

    if amp:
        if as_pandas:
            out = pd.DataFrame({"freq":peak_freq, "amp":amp_peak_freq}, index=[0])
        else:
            out = np.array([peak_freq, amp_peak_freq])
    else:
        if as_pandas:
            out = pd.Series([peak_freq], index=["peak_freq"])
        else:
            out = peak_freq
    
    return out

#%%
def spectral_quantile(s, fs, q=[0.05, 0.25, 0.5, 0.75, 0.95], nperseg=1024, roi=None,
                      as_pandas=False, amp=False, **kwargs):
    """
    Compute the q-th quantile of the power spectrum. If a region of interest with time and
    spectral limits is provided, the q-th quantile is computed on the selection.

    Parameters
    ----------
    s : 1D array
        Input audio signal
    fs : float
        Sampling frequency of audio signal
    q : array or float, optional
        Quantile or sequence of quantiles to compute, which must be between 0 and 1
        inclusive.
        The defaul is [0.05, 0.25, 0.5, 0.75, 0.95].
    nperseg : int, optional
        Length of segment to compute the FFT. The default is 1024.
    roi : pandas.Series, optional
        Region of interest where peak frequency will be computed.
        Series must have a valid input format with index: min_t, min_f, max_t, max_f.
        The default is None.
    as_pandas: bool
        Return data as a pandas.Series or pandas.DataFrame, when amp
        is False or True, respectively. Default is False.
    amp: bool
        Enable quantiles amplitude output. Default is False.
    kwargs : additional keyword arguments
        If `window='hann'`, additional keyword arguments to pass to
        `sound.spectrum`.
        
    Returns
    -------
    Pandas Series or Numpy array
        Quantiles of power spectrum.

    Examples
    --------
    >>> from maad import sound, features
    >>> s, fs = sound.load('../data/spinetail.wav')

    Compute the q-th quantile of the power spectrum

    >>> qs = features.spectral_quantile(s, fs, [0.05, 0.25, 0.5, 0.75, 0.95], as_pandas=True)
    >>> print(qs)
    0.05    6029.296875
    0.25    6416.894531
    0.50    6632.226562
    0.75    6890.625000
    0.95    9216.210938
    dtype: float64

    Compute the q-th quantile of the power spectrum and its amplitude

    >>> qs = features.spectral_quantile(s, fs, [0.05, 0.25, 0.5, 0.75, 0.95], amp=True, as_pandas=True)
    >>> print(qs)
                 freq       amp
    0.05  6029.296875  0.000007
    0.25  6416.894531  0.000067
    0.50  6632.226562  0.000087
    0.75  6890.625000  0.000022
    0.95  9216.210938  0.000004

    """
    q = np.asanyarray(q)
    if not _quantile_is_valid(q):
        raise ValueError("Percentiles must be in the range [0, 1]")

    # Compute spectrum
    # if roi is None:
    #     pxx, fidx = sound.spectrum(s, fs, nperseg, **kwargs)
    # else:
    #     pxx, fidx = sound.spectrum(s, fs, nperseg,
    #                                tlims=[roi.min_t, roi.max_t],
    #                                flims=[roi.min_f, roi.max_f],
    #                                **kwargs)
    # pxx = pd.Series(pxx, index=fidx)
    if roi is None:
        Sxx,_,fn,_ = sound.spectrogram(s, fs, nperseg=nperseg, **kwargs)
    else:
        Sxx,_,fn,_ = sound.spectrogram(s, fs, nperseg=nperseg,
                                       tlims=[roi.min_t, roi.max_t],
                                       flims=[roi.min_f, roi.max_f],
                                        **kwargs)
    pxx = pd.Series(np.average(Sxx,axis=1), index=fn)
    
    # Compute spectral q
    norm_cumsum = pxx.cumsum()/pxx.sum()
    spec_quantile = []
    for quantile in q:
        spec_quantile.append(pxx.index[np.where(norm_cumsum>=quantile)[0][0]])

    if amp:
        if as_pandas:
            out = pd.DataFrame({"freq":spec_quantile, "amp":pxx[spec_quantile].values}, index=q)
        else:
            out = np.transpose(np.array([q, spec_quantile, pxx[spec_quantile]]))
    else:
        if as_pandas:
            out = pd.Series(spec_quantile, index=q)
        else:
            out = np.array(spec_quantile)

    return out

#%%
def spectral_bandwidth(s, fs, nperseg=1024, roi=None, as_pandas=False, **kwargs):
    """
    Compute the bandwith of the power spectrum. If a region of interest with time
    and spectral limits is provided, the spectral bandwidth is computed on the selection.

    Parameters
    ----------
    s : 1D array
        Input audio signal
    fs : float
        Sampling frequency of audio signal
    nperseg : int, optional
        Length of segment to compute the FFT. The default is 1024.
    roi : pandas.Series, optional
        Region of interest where peak frequency will be computed.
        Series must have a valid input format with index: min_t, min_f, max_t, max_f.
        The default is None.
    as_pandas : bool
        Return data as a pandas.Series. This is usefull when computing multiple features
        over a signal. Default is False.
    kwargs : additional keyword arguments
        If `window='hann'`, additional keyword arguments to pass to
        `sound.spectrum`.

    Returns
    -------
    bandwidth_50 : float
        Bandwidth 50% of the audio
    bandwidth_90 : float
        Bandwidth 90%  of the audio
    
    Examples
    --------
    >>> from maad import features, sound
    >>> s, fs = sound.load('../data/spinetail.wav')
    >>> bw_50, bw_90 = features.spectral_bandwidth(s, fs, nperseg=1024)
    >>> print("Bandwidth 50% : {:.4f} / Bandwidth 90% : {:.4f}".format(bw_50, bw_90))
    Bandwidth 50% : 473.7305 / Bandwidth 90% : 3186.9141

    """
    # Compute spectral bandwith with the quantiles
    q = spectral_quantile(
        s, fs, [0.05, 0.25, 0.75, 0.95], nperseg, roi, False, True, **kwargs)
    
    out = np.array([np.abs(q[2,1]-q[1,1]), np.abs(q[3,1]-q[0,1])])

    if as_pandas:
        out = pd.Series(out, index=["bw_50", "bw_90"])
        
    return out

#%%
def all_spectral_features(s, fs, nperseg=1024, roi=None, method='fast', **kwargs):
    """
    Compute all the spectral features for a signal.

    Parameters
    ----------
    s : 1D array
        Input audio signal
    fs : float
        Sampling frequency of audio signal
    nperseg : int, optional
        Length of segment to compute the FFT. The default is 1024.
    roi : pandas.Series, optional
        Region of interest where peak frequency will be computed.
        Series must have a valid input format with index: min_t, min_f, max_t, max_f.
        The default is None.
    method : {'fast', 'best'}, optional
        Method used to compute the peak frequency.
        The default is 'fast'.
    kwargs : additional keyword arguments
        If `window='hann'`, additional keyword arguments to pass to
        `sound.spectrum`.

    Returns
    -------
    spectral_features : pandas DataFrame
        DataFrame with all spectral features computed in the spectrum
    
    Examples
    --------
    >>> from maad import features, sound
    >>> s, fs = sound.load('../data/spinetail.wav')

    Compute all the spectral features

    >>> features.all_spectral_features(s, fs, nperseg=1024, roi=None)
    sm           2.276330e-06
    sv           8.118042e-11
    ss           5.844664e+00
    sk           4.048891e+01
    freq_05      6.029297e+03
    freq_25      6.416895e+03
    freq_50      6.632227e+03
    freq_75      6.890625e+03
    freq_95      9.216211e+03
    peak_freq    6.632227e+03
    bw_50        4.737305e+02
    bw_90        3.186914e+03
    dtype: float64
    """

    # Compute transformations
    Sxx_power,_,_,_ = sound.spectrogram (s, fs, window='hann', nperseg=nperseg)
    S_power = sound.avg_power_spectro(Sxx_power)
    
    # Compute features
    sm = spectral_moments(S_power)
    qs = spectral_quantile(s, fs, [0.05, 0.25, 0.5, 0.75, 0.95], nperseg, roi, **kwargs)
    bw_50, bw_90 = spectral_bandwidth(s, fs, nperseg, roi, **kwargs)
    peak_freq = peak_frequency(s, fs, method, nperseg, roi, **kwargs)

    # Organize data into a Dataframe
    spectral_features = pd.Series({
        "sm":sm[0], 
        "sv":sm[1], 
        "ss":sm[2], 
        "sk":sm[3],
        "freq_05":qs[0], 
        "freq_25":qs[1], 
        "freq_50":qs[2], 
        "freq_75":qs[3], 
        "freq_95":qs[4], 
        "peak_freq":peak_freq,
        "bw_50":bw_50, 
        "bw_90":bw_90})

    return spectral_features

if __name__ == "__main__":
    import doctest
    doctest.testmod()


""" 
Utility functions for computing composite soundscape descriptors using data from multiple files at the same sampling site.
"""
import os
from concurrent.futures import ProcessPoolExecutor
import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
from maad import sound, util, rois

#%% Function argument validation
def _input_validation(data_input, verbose=True):
    """ Validate dataframe or path input argument """
    if isinstance(data_input, pd.DataFrame):
        df = data_input

    elif isinstance(data_input, str):
        if os.path.isdir(data_input):
            if verbose:
                print('Collecting metadata from directory path...')
            df = util.get_metadata_dir(data_input)
            df['time'] = df.date.dt.hour
            if verbose:
                print('Done!')
        elif os.path.isfile(data_input) and data_input.lower().endswith(".csv"):
            if verbose:
                print('Loading metadata from csv file')
            try:
                # Attempt to read all wav data from the provided file path.
                df = pd.read_csv(data_input, dtype={'time': str}) 
                df['date'] = pd.to_datetime(df.date)
                df['time'] = df.date.dt.hour
            except FileNotFoundError:
                raise FileNotFoundError(f"File not found: {data_input}")
            if verbose:
                print('Done!')
    
    else:
        raise ValueError("Input 'data' must be either a Pandas DataFrame or a file path string")
    
    return df

def _validate_n_jobs(n_jobs, verbose=True):
    """ Validate number of jobs """
    if not isinstance(n_jobs, int):
        raise ValueError("n_jobs must be an integer.")
    
    if n_jobs <= 0 or n_jobs == -1:
        # Use all available CPUs
        n_jobs = os.cpu_count()
        if verbose:
            print(f"Using all available CPUs: {n_jobs}")
    else:
        # Validate that the number is not larger than available CPUs
        available_cpus = os.cpu_count()
        if n_jobs > available_cpus:
            # Set n_jobs to the maximum number of available CPUs
            n_jobs = available_cpus
            if verbose:
                print(f"Adjusted n_jobs to maximum available CPUs: {n_jobs}")
        else:
            if verbose:
                print(f"Using specified number of CPUs: {n_jobs}")

    return n_jobs

#%%
def _spectral_peak_density(
        path_audio, target_fs, nperseg, noverlap, db_range, min_distance, threshold_abs, verbose=True):
    """
    Computes the spectral peak density for an audio file, representing the number of peaks per time step within each frequency bin.

    Parameters
    ----------
    path_audio : pandas DataFrame
        A Pandas DataFrame containing information about the audio files.
    target_fs : int
        The target sample rate to resample the audio signal if needed.
    nperseg : int
        Window length of each segment to compute the spectrogram.
    noverlap : int
        Number of samples to overlap between segments to compute the spectrogram.
    db_range : float
        Dynamic range of the computed spectrogram.
    min_distance : int
        Minimum number of indices separating peaks.
    threshold_abs : float
        Minimum amplitude threshold for peak detection in decibels.

    Returns
    -------
    peak_density : pandas Dataframe
        The peak density representation of the audio per frequency bin.

    """

    filename = os.path.basename(path_audio)
    if verbose:
        print(f'Processing file {filename}', end='\r')

    # Load data
    s, fs = sound.load(path_audio)
    s = sound.resample(s, fs, target_fs, res_type="scipy_poly")
    Sxx, tn, fn, ext = sound.spectrogram(s, target_fs, nperseg=nperseg, noverlap=noverlap)
    Sxx_db = util.power2dB(Sxx, db_range=db_range)

    # Compute local max
    _, peak_freq = rois.spectrogram_local_max(
        Sxx_db, tn, fn, ext,
        min_distance, 
        threshold_abs)

    # Compute peak density (number of peaks / time steps)
    freq_idx, count_freq = np.unique(peak_freq, return_counts=True)
    count_peak = np.zeros(fn.shape)
    bool_index = np.isin(fn, freq_idx)
    indices = np.where(bool_index)[0]
    count_peak[indices] = count_freq / len(tn)
    peak_density = pd.Series(index=fn, data=count_peak)

    # Normalize per time
    peak_density.name = filename
    return peak_density.to_frame().T

#%%
def graphical_soundscape(
    data, threshold_abs, path_audio='path_audio', time='time', target_fs=48000, nperseg=256, noverlap=128, db_range=80, min_distance=1, n_jobs=1, verbose=True):
    """
    Computes a graphical soundscape from a given DataFrame of audio files.

    This function is a variant of the original graphical soundscapes introduced by Campos-Cerqueira et al. The peaks are detected on the spectrogram instead of detecting peaks on the spectrum. Results are similar but not equal to the ones computed using seewave in R.

    References:
        - Campos‐Cerqueira, M., et al., 2020. How does FSC forest certification affect the acoustically active fauna in Madre de Dios, Peru? Remote Sensing in Ecology and Conservation 6, 274–285. https://doi.org/10.1002/rse2.120
        - Furumo, P.R., Aide, T.M., 2019. Using soundscapes to assess biodiversity in Neotropical oil palm landscapes. Landscape Ecology 34, 911–923.
        - Campos-Cerqueira, M., Aide, T.M., 2017. Changes in the acoustic structure and composition along a tropical elevational gradient. JEA 1, 1–1. https://doi.org/10.22261/JEA.PNCO7I

    Parameters
    ----------
    data : pandas DataFrame
        A Pandas DataFrame containing information about the audio files.
        If a string is passed with a directory location or a csv file, parameters 'path_audio' and 'time' will be set as default and can't be customized.
    threshold_abs : float
        Minimum amplitude threshold for peak detection in decibels.
    path_audio : str
        Column name where the full path of audio is provided.
    time : str
        Column name where the time is provided as a string using the format 'HHMMSS'.
    target_fs : int
        The target sample rate to resample the audio signal if needed.
    nperseg : int
        Window length of each segment to compute the spectrogram.
    noverlap : int
        Number of samples to overlap between segments to compute the spectrogram.
    db_range : float
        Dynamic range of the computed spectrogram.
    min_distance : int
        Minimum number of indices separating peaks.
    n_jobs : int
        Number of processes to use for parallel computing. Default is 1.
    verbose : bool
        Whether to print messages. Default is True.

    Returns
    -------
    res : pandas DataFrame
        A Pandas DataFrame containing the graphical representation of the soundscape.

    Examples
    --------
    >>> from maad.util import get_metadata_dir
    >>> from maad.features import graphical_soundscape, plot_graph
    >>> df = get_metadata_dir('../../data/indices')
    >>> df['hour'] = df.date.dt.hour
    >>> gs = graphical_soundscape(data=df, threshold_abs=-80, time='hour')
    >>> plot_graph(gs)
    """
    df = _input_validation(data, verbose=verbose)
    df.sort_values(by=path_audio, inplace=True)
    total_files = len(df)
    if verbose:
        print(f'{total_files} files found to process...')
    flist = df[path_audio].to_list()

    if n_jobs == 1:
        # Use sequential processing
        results = []
        for path_audio in flist:
            result = _spectral_peak_density(path_audio, target_fs, nperseg, noverlap, db_range, min_distance, threshold_abs, verbose=verbose)
            results.append(result)
    else:
        # Use parallel processing
        _validate_n_jobs(n_jobs, verbose=verbose)
        with ProcessPoolExecutor() as executor:
            futures = [executor.submit(_spectral_peak_density, path_audio, target_fs, nperseg, noverlap, db_range, min_distance, threshold_abs, verbose)
                       for path_audio in flist]

        # Wait for all tasks to complete
        results = [future.result() for future in futures]
        
    res = pd.concat(results)
    res['time'] = df[time].values
    if verbose:
        print('\nComputation completed!')
    return res.groupby('time').mean()

#%%
def plot_graph(graph, ax=None, savefig=False, fname=None):
    """ Plots a graphical soundscape

    Parameters
    ----------
    graph : pandas.Dataframe
        A graphical soundscape as pandas dataframe with index as time and frequency as columns
    ax : matplotlib.axes, optional
        Axes for subplots. If not provided it creates a new figure, by default None.

    Returns
    -------
    ax
        Axes of the figure
    """
    if ax == None:
        fig, ax = plt.subplots()

    ax.imshow(graph.values.T, aspect='auto', origin='lower', 
              extent=[int(graph.index[0]), int(graph.index[-1]), 
                      float(graph.columns[0]), float(graph.columns[-1])])
    ax.set_xlabel('Time (h)')
    ax.set_ylabel('Frequency (Hz)')

    if savefig:
        plt.savefig(fname, bbox_inches='tight')
    
    return ax



#!/usr/bin/env python
""" 
Collection of functions to transform audio signals : Take the envelope
"""
#
# Authors:  Juan Sebastian ULLOA <lisofomia@gmail.com>
#           Sylvain HAUPERT <sylvain.haupert@mnhn.fr>
#
# License: New BSD License

# =============================================================================
# Load the modules
# =============================================================================
# Import external modules
import numpy as np
import pandas as pd
from scipy.signal import hilbert
from scipy.signal import periodogram, welch
import scipy
from warnings import warn

# import internal modules
from maad.sound import wave2frames
from maad.util import plot_spectrum

#%%
# =============================================================================
# public functions
# =============================================================================
def envelope (s, mode='fast', Nt=32):
    """
    Calculate the envelope of a sound waveform (1d)
    
    Parameters
    ----------
    s : 1d ndarray of floats 
        Vector containing sound waveform 
    mode : str, optional, default is `fast`
        - `fast` : The sound is first divided into frames (2d) using the 
            function wave2timeframes(s), then the max of each frame gives a 
            good approximation of the envelope.
        - `Hilbert` : estimation of the envelope from the Hilbert transform. 
            The method is slow
    Nt : integer, optional, default is `32`
        Size of each frame. The largest, the highest is the approximation.
                  
    Returns
    -------    
    env : 1d ndarray of floats
        Envelope of the sound
        
    References
    ----------
    .. [1] Towsey, Michael (2013), Noise Removal from Waveforms and Spectrograms Derived from Natural Recordings of the Environment. Queensland University of Technology, Brisbane.
    .. [2] Towsey, Michael (2017),The calculation of acoustic indices derived from long-duration recordings of the natural environment. Queensland University of Technology, Brisbane.
    
    Examples
    --------
    >>> s,fs = maad.sound.load("../data/rock_savanna.wav")
    >>> env_fast = maad.sound.envelope(s, mode='fast', Nt=32)
    >>> env_fast
    array([0.2300415 , 0.28643799, 0.24285889, ..., 0.3059082 , 0.20040894,
       0.26074219])
    
    >>> env_hilbert = maad.sound.envelope(s, mode='hilbert')
    >>> env_hilbert
    array([0.06588196, 0.11301711, 0.09201435, ..., 0.18053983, 0.18351906,
       0.10258595])
    
    Compute the time vector for the vector wave.
    
    >>> import numpy as np
    >>> t = np.arange(0,len(s),1)/fs
    
    Compute the time vector for the vector env_fast.
    
    >>> t_env_fast = np.arange(0,len(env_fast),1)*len(s)/fs/len(env_fast)
    
    Plot 0.1s of the envelope and 0.1s of the abs(s).
    
    >>> import matplotlib.pyplot as plt
    >>> fig1, ax1 = plt.subplots(figsize=(10,4))
    >>> ax1.plot(t[t<0.1], abs(s[t<0.1]), label='abs(s)', lw=0.7)
    >>> ax1.plot(t[t<0.1], env_hilbert[t<0.1], label='env(s) - hilbert option', lw=0.7)
    >>> ax1.plot(t_env_fast[t_env_fast<0.1], env_fast[t_env_fast<0.1], label='env(s) - fast option', lw=0.7)
    >>> ax1.set_xlabel('Time [sec]')
    >>> ax1.set_ylabel('Amplitude')
    >>> ax1.legend()
    """
    if mode == 'fast' :
        # Envelope : take the max (see M. Towsey) of each frame
        frames = wave2frames(s, Nt)
        env = np.max(abs(frames),0) 
    elif mode =='hilbert' :
        # Compute the hilbert transform of the waveform and take the norm 
        # (magnitude) 
        env = np.abs(hilbert(s))  
    else:
        raise Exception("Invalid mode. Mode should be 'fast' or 'hilbert' ")
        
    return env

#%%
def spectrum(s, fs, nperseg=256, noverlap=None, nfft=None, window='hann', method='welch', 
        tlims=None, flims=None, scaling='spectrum', as_pandas_series=False, display=False):
    """ 
    Estimate the power spectral density or power spectrum of 1D signal.
    
    The estimates can be computed using two methods: Welch or periodogram. Welch's method
    divides the signal into segments, computes the power spectral density for each 
    segment and then takes the average between segments. The periodogram method computes
    the power spectral density of the input signal using a defined window 
    (Hanning window by default).
    
    Parameters
    ----------
    s: 1D array 
        Input signal to process 
    
    fs: float, optional
        Sampling frequency of audio signal
    
    nperseg: int, optional
        Length of segment for 'welch' method, default is 256
    
    noverlap: int, optional
        Overlap between segments for Welch's method. If None, noverlap = nperseg/2.
    
    window : string, default is 'hann'
        Name of the window used for the short fourier transform, 'hanning' window is 
        now deprecated.
    
    nfft: int, optional
        Length of FFT for periodogram method. If None, length of signal will be used.
        Length of FFT for welch method if zero padding is desired. If None, length of nperseg will be used.
    
    method: {'welch', 'periodogram'}
        Method used to estimate the power spectral density of the signal
    
    tlims: tuple of ints or floats
        Temporal limits to compute the power spectral density in seconds (s)
        If None, estimates for the complete signal will be computed.
        Default is 'None'
    
    flims: tuple of ints or floats
        Spectral limits to compute the power spectral density in Hertz (Hz)
        If None, estimates from 0 to fs/2 will be computed.
        Default is 'None'
        
    scaling: {'spectrum', 'density'}, optional
        Choose between power spectrum (units V**2) or power spectral density (units V**2/Hz) scaling.
        Defaults to 'spectrum'.
    
    as_pandas_series: bool
        Return data as a pandas.Series. This is usefull when computing multiple features
        over a signal. Default is False.
        

    
    Returns
    -------
    pxx: pandas Series
        Power spectral density estimate.
    
    f_idx: pandas Series
        Index of sample frequencies.
    
    Notes
    -----
    This is a wrapper that uses functions from Scipy (scipy.org), in particular from the scipy.signal module.
    
    Examples
    --------
    >>> from maad import sound
    >>> s, fs = sound.load('../data/spinetail.wav')
    >>> spec, f_idx = sound.spectrum(s, fs, nperseg=512, display=True)
    
    Specify temporal and spectral limits to get spectral characteristics of the 
    spinetail's song.
    
    >>> spec, f_idx = sound.spectrum(s, fs, nperseg=2056, tlims=(5.3, 7.9), flims=(2000, 12000), display=True)
    
    """
    if tlims is not None:
    # trim audio signal
        try:
            s = s[int(tlims[0]*fs): int(tlims[1]*fs)]
        except:
            raise Exception('length of tlims tuple should be 2')
    
    # small fix to keep compatibility with old scipy version
    if window=='hanning':
        window='hann'
        warn("Window type 'hanning' will be deprecated. Changing to 'hann' window.", FutureWarning)

    if method=='welch':
        f_idx, pxx = welch(s, fs, window, nperseg, noverlap, nfft, scaling=scaling)
    
    elif method=='periodogram':
        f_idx, pxx = periodogram(s, fs, window, nfft, scaling=scaling)
        
    else:
        raise Exception("Invalid method. Method should be 'welch' or 'periodogram' ")
        
    if flims is not None:
        pxx = pxx[(f_idx>=flims[0]) & (f_idx<=flims[1])]
        f_idx = f_idx[(f_idx>=flims[0]) & (f_idx<=flims[1])]

    if as_pandas_series==True:
        index_names = ['psd_' + str(idx).zfill(3) for idx in range(1,len(pxx)+1)]
        pxx = pd.Series(pxx, index=index_names)
        f_idx = pd.Series(f_idx, index=index_names)
        
    if display:
        plot_spectrum(pxx, f_idx)

    return pxx, f_idx

#%%
def resample(s, fs, target_fs, res_type = 'scipy_poly', **kwargs):
    """
    Changes the sample rate of an audio file or any time series.

    Parameters
    ----------
    s : np.ndarray 
        Mono or stereo signal as NumPy array.

    fs : int
        Time series sampling rate. 

    target_fs : int
        Target sampling rate.

    res_type : str, optional
        Resample method. By default 'kaiser_best'.
        'kaiser_best' method that applies a Kaiser windowed sinc interpolation.
        'kaiser_fast' similar to kaiser_best, but faster.
        'scipy' is a Fourier transform method implemented in scipy.
        'scipy_poly' Resample using polyphase filtering and an FIR filter. 

    kwargs : additional keyword arguments
        If `res_type='scipy'`, additional keyword arguments to pass to
        `scipy.signal.resample`.

    Returns
    -------
    res_data : np.ndarray 
        Resampling time series.

    See Also
    --------
    scipy.signal.resample
    
    Examples
    --------
    
    Resample an audio file from sample rate of 44100 kHz to 22050.
    
    >>> from maad import sound
    >>> s, fs = sound.load('../data/spinetail.wav')
    >>> s_resamp = sound.resample(s, fs, target_fs=fs/2)
    >>> print('Number of samples - original audio:', s.shape[0],
    ...       'Number of samples - resampled audio:', s_resamp.shape[0])
    >>> _ = sound.spectrogram(s, fs, display=True)
    >>> _ = sound.spectrogram(s_resamp, fs/2, display=True)

    """
    if fs == target_fs:
        return s

    if res_type == 'scipy':
        # Calculate the resampling ratio and get the number of output samples
        ratio = float(target_fs) / fs
        n_samples = int(np.round(s.shape[-1] * ratio))
        res_data = scipy.signal.resample(s, n_samples, axis=-1, **kwargs)
    
    elif res_type == 'scipy_poly':
        # Calculate the resampling ratio and get the common denominator
        gcd = np.gcd(int(fs), int(target_fs))
        num = target_fs // gcd
        den = fs // gcd
        # Resample the signal using SciPy's resample_poly function
        res_data = scipy.signal.resample_poly(s, num, den)
    
    elif res_type in ('kaiser_best', 'kaiser_fast'):
        raise ValueError("'kaiser_best' and 'kaiser_fast' methods are deprecated. Valid resample methods are 'scipy', 'scipy_poly', not res_type={}".format(res_type))
    
    else:
        raise ValueError("Valid resample methods are 'scipy', 'scipy_poly', not res_type={}".format(res_type))
                         
    return np.ascontiguousarray(res_data, dtype=s.dtype)

#%%
def trim(s, fs, min_t, max_t, pad=False, pad_constant=0):
    """
    Slices a time series, from a initial time `min_t` to an ending time `max_t`.  
    If the target duration `duration =` is larger than the original duration and `pad = True`, 
    the time series is padded with a constant value `pad_constant = 0`.

    Parameters
    ----------
    s : np.ndarray 
        Mono or stereo signal as NumPy array.

    fs : int
        Time series sampling rate.

    min_t : float
        Initial time. If initial time `min_t < 0` and `pad=True`, this time is added to the 
        beginning of the audio slice.

    max_t : float
        Ending time of the audio slice.
        
    pad : bool, optional
        If true, the time series is padded with a constant value `pad_constant`. Default is False.
        
    pad_constant : 
        It is the constant with which the time series is padded. 
        
    Returns
    -------
    s_slice : np.ndarray 
        Time series with duration `duration = max_t - min_t`.

    See Also
    --------
    numpy.pad
    
    Examples
    --------
    
    Slice an audio file from 5 to 8 seconds.
    
    >>> from maad import sound
    >>> s, fs = sound.load('../data/spinetail.wav') 
    >>> s_slice = sound.trim(s, fs, min_t = 5, max_t = 8)
    >>> _ = sound.spectrogram(s_slice, fs, display=True, figsize=(4,6))
    >>> s_slice.shape[0]/fs
    3.0
    """

    min_lim = min_t * fs
    max_lim = max_t * fs
    duration = max_t - min_t
    
    if pad:
        
        if s.ndim == 1:  # mono file
            
            if (duration * fs) > s.shape[-1]:
                # length of trimmed section larger than signal duration, pad both sides
                up_lim = int(max_t * fs - s.shape[-1])
                if min_t <=0:
                    low_lim = int(abs(min_t) * fs) 
                    s_slice = np.pad(s, (low_lim, up_lim), 'constant',
                                     constant_values=(pad_constant))
                else:
                    low_lim = int(0.0) 
                    s_slice = np.pad(s, (low_lim, up_lim), 'constant',
                                     constant_values=(pad_constant))
                    s_slice = s_slice[int(min_lim):]

            elif min_t < 0:
                # min_t with negative values, pad left side
                pad_width = int(abs(min_t) * fs)
                up_lim = int(max_t * fs)
                s_slice = s[0:up_lim]
                s_slice = np.pad(s_slice, (pad_width, 0), 'constant', 
                                 constant_values=(pad_constant))

            elif max_t * fs > s.shape[-1]:
                # max_t with values outside duration of signal, pad right side
                low_lim = int(min_t * fs)
                pad_width = int((max_t * fs) - s.shape[-1])
                s_slice = s[low_lim:]
                s_slice = np.pad(s_slice, (0, pad_width), 'constant', 
                                 constant_values=(pad_constant))
                                 
            else:
                # no padding needed
                s_slice = s[int(min_lim): int(max_lim)]
        
        else:  # stereo file
            
            s_slice = np.empty((s.shape[0], int(duration * fs)))
            for i, chanel in enumerate(s):
                if (duration * fs) > chanel.shape[-1]:
                    up_lim = int(max_t * fs - chanel.shape[-1])
                    
                    if min_t <=0:
                        low_lim = int(abs(min_t) * fs) 
                        s_slice = np.pad(chanel, (low_lim, up_lim), 'constant',
                                         constant_values=(pad_constant))
                    else:
                        low_lim = int(0.0) 
                        s_slice = np.pad(chanel, (low_lim, up_lim), 'constant',
                                         constant_values=(pad_constant))
                        s_slice = s_slice[int(min_lim):]
                else:
                    s_slice[i] = s[i][int(min_lim): int(max_lim)]
                    
        return s_slice
    
    else:
        
        if min_t < 0:
            raise ValueError("min_t must be >= 0.")
        
        if s.ndim == 1:
            if (duration * fs) > s.shape[-1]:
                raise ValueError("Target duration is longer than original duration.")
            else:
                s_slice = s[int(min_lim): int(max_lim)]
        else:   
            
            s_slice = np.empty((s.shape[0], int(duration * fs)))
            for i, chanel in enumerate(s):
                if (duration * fs) > chanel.shape[-1]:
                    raise ValueError("Target duration is longer than original duration.")
                else:
                    s_slice[i] = s[i][int(min_lim): int(max_lim)]                
        
        return s_slice
    
def normalize(s, max_amp=0.7, max_db=None):
    """
    Normalize audio signal to desired amplitude or decibell full scale value (dBFS).
    
    Parameters
    ----------
    s : np.ndarray 
        Mono audio signal as NumPy array.
    max_amp : float, optional
        Maximum amplitude. Must be between 0-1. The default is 0.7.
    max_db : float, optional
        Maximum amplitude in dBFS. Must be a negative value.

    Returns
    -------
    s_out : np.ndarray
        Normalized mono signal.
        
    See Also
    --------
    sound.trim
    
    Examples
    --------
    >>> import numpy as np
    >>> from maad import sound
    
    Use normalized amplitude value for normalization
    
    >>> s, fs = sound.load('../data/spinetail.wav')
    >>> s_out = sound.normalize(s, max_amp=0.8)
    >>> print('s_out maximum amplitude:', np.abs(s_out).max())
    s_out maximum amplitude: 0.8
    
    Use a dBFS value for normalization:
    
    >>> s, fs = sound.load('../data/spinetail.wav')
    >>> s_out = sound.normalize(s, max_db=-3)
    >>> print('s_out maximum value in dBFS:', np.max(20*np.log10(np.abs(s_out))))
    s_out maximum value in dBFS: -3.0
    
    """
    # Check values for normalization and transform if using max_dbfs
    if (max_amp < 0) | (max_amp > 1):
        raise ValueError('max_amp should be a value between 0 and 1')
        
    elif (max_db is not None):
        if max_db > 0:
            raise ValueError('max_db should be a negative value')
        else:
            # transform dbfs to normalized amplitude
            max_amp = 10 ** (max_db/20)
    else:
        pass
        
    # compute and apply gain value
    gain = max_amp/np.abs(s).max()
    s_out = s*gain
    
    return s_out

def gain(s, gain_db=0):
    """
    Amply amplification or attenuation to the audio signal.
    
    Parameters
    ----------
    s : np.ndarray 
        Mono audio signal as NumPy array.
    gain_db : float, optional
        Gain value to amplify (positive values), or attenuate (negative values) the 
        signal in dB.

    Returns
    -------
    s_out : np.ndarray
        Amplified or attenuated audio signal.
        
    See Also
    --------
    sound.normalize, sound.trim
    
    Examples
    --------
    >>> import numpy as np
    >>> from maad import sound
    
    Amplify the signal by 3 dB
    
    >>> s, fs = sound.load('../data/spinetail.wav')
    >>> s_out = sound.gain(s, gain_db=3)
    >>> print('Signal amplified by:', 20*np.log10((s_out.max()/s.max())), 'dB')
    Signal amplified by: 3 dB
    
    Attenuate the signal by 3 dB
    
    >>> s, fs = sound.load('../data/spinetail.wav')
    >>> s_out = sound.gain(s, gain_db=-3)
    >>> print('Signal amplified by:', 20*np.log10((s_out.max()/s.max())), 'dB')
    Signal amplified by: -3 dB
    
    """        
    # Apply the gain value
    
    s_out = s * 10**(gain_db/20)
    
    return s_out


#!/usr/bin/env python 
"""  
Segmentation methods to find regions of interest in the time and frequency domain.
""" 
#
# Authors:  Juan Sebastian ULLOA <lisofomia@gmail.com>
#           Sylvain HAUPERT <sylvain.haupert@mnhn.fr>
#
# License: New BSD License

# =============================================================================
# Load the modules
# =============================================================================
# Import external modules 
import numpy as np  
from scipy.stats import iqr 
from skimage import measure
from skimage.feature import peak_local_max
import pandas as pd 
import sys 
import matplotlib.pyplot as plt
_MIN_ = sys.float_info.min 
 
# Import internal modules 
from maad.util import (plot2d, rand_cmap, plot_spectrogram)

#%%
#**************************************************************************** 
# private functions                
#**************************************************************************** 
def _double_threshold_rel (im, bin_std=6, bin_per=0.5, 
                           verbose=False, display=False, savefig=None, **kwargs): 
    """ 
    Binarize an image based on a double relative threshold.  
    The values used for the thresholding depends on the values found in the  
    image. => relative threshold 
     
    Parameters 
    ---------- 
    im : 2d ndarray of scalars 
        Spectrogram (or image) 
 
    bin_std : scalar, optional, default is 6 
        Set the first threshold. This threshold is not an absolute value but 
        depends on values that are similar to 75th percentile (pseudo_mean) and 
        a sort of std value of the image.  
        threshold1 = "pseudo_mean" + "std" * bin_std    
        Value higher than threshold1 are set to 1, they are the seeds for  
        the second step. The others are set to 0.  
         
    bin_per: scalar, optional, defautl is 0.5 
        Set how much the second threshold is lower than the first 
        threshold value. From 0 to 1. ex: 0.1 = 10 %. 
        threshold2 = threshold1 (1-bin_per)   
        Value higher than threshold2 and connected (directly or not) to the  
        seeds are set to 1, the other remains 0 
 
    verbose : boolean, optional, default is False
        print messages
    
    display : boolean, optional, default is False 
        Display the signal if True 
         
    savefig : string, optional, default is None 
        Root filename (with full path) is required to save the figures. Postfix 
        is added to the root filename. 
         
    kwargs, optional. This parameter is used by plt.plot and savefig functions 
             
        - savefilename : str, optional, default :'_spectro_after_noise_subtraction.png' 
            Postfix of the figure filename 
          
        - figsize : tuple of integers, optional, default: (4,10) 
            width, height in inches.   
         
        - title : string, optional, default : 'Spectrogram' 
            title of the figure 
         
        - xlabel : string, optional, default : 'Time [s]' 
            label of the horizontal axis 
         
        - ylabel : string, optional, default : 'Amplitude [AU]' 
            label of the vertical axis 
         
        - cmap : string or Colormap object, optional, default is 'gray' 
            See https://matplotlib.org/examples/color/colormaps_reference.html 
            in order to get all the  existing colormaps 
            examples: 'hsv', 'hot', 'bone', 'tab20c', 'jet', 'seismic',  
            'viridis'... 
         
        - vmin, vmax : scalar, optional, default: None 
            `vmin` and `vmax` are used in conjunction with norm to normalize 
            luminance data.  Note if you pass a `norm` instance, your 
            settings for `vmin` and `vmax` will be ignored. 
         
        - extent : scalars (left, right, bottom, top), optional, default: None 
            The location, in data-coordinates, of the lower-left and 
            upper-right corners. If `None`, the image is positioned such that 
            the pixel centers fall on zero-based (row, column) indices. 
         
        - dpi : integer, optional, default is 96 
            Dot per inch.  
            For printed version, choose high dpi (i.e. dpi=300) => slow 
            For screen version, choose low dpi (i.e. dpi=96) => fast 
         
        - format : string, optional, default is 'png' 
            Format to save the figure 
             
        ... and more, see matplotlib    
     
    Returns 
    ------- 
    im_out: binary image  
     
    References 
    ---------- 
    .. [1] from MATLAB: Threshold estimation (from Oliveira et al, 2015) 
       Adapted by S. Haupert Dec 12, 2017 
    """ 
    
    # test if im is full of zeros
    if not im.any() :
        im_out = np.zeros(im.shape)   
    else:
    
        # Compute the qth percentile of the data along the specified axis 
        val1 = np.percentile(im[np.where(im>0)],75)   # value corresponding to the limit between the 75% lowest value and 25% largest value 
         
        # The interquartile range (IQR) is the difference between the 75th and  
        # 25th percentile of the data. It is a measure of the dispersion similar  
        # to standard deviation or variance, but is much more robust against outliers 
        val2 = iqr(im[np.where(im>0)])*bin_std 
         
        # Threshold : qth  percentile + sort of std 
        h_th = val1 + val2 
        # Low threshold limit 
        l_th = (h_th-h_th*bin_per) 
        
        if verbose :
            print(72 * '_') 
            print('Double thresholding with values relative to the image...') 
            print ('**********************************************************') 
            print ('  high threshold value %.2f | low threshold value %.2f' % (h_th, l_th)) 
            print ('**********************************************************') 
         
        # binarisation  
        im_t1 = im > h_th    # mask1 
        im_t2 = im > l_th    # mask2 
        im_t3 = im * im_t1   # selected parts of the image 
         
        #find index of regions which meet the criteria 
        conncomp_t2 = measure.label(im_t2)    #Find connected components in binary image 
        rprops = measure.regionprops(conncomp_t2,im_t3)  
         
        rprops_mean_intensity = [region['mean_intensity'] for region in rprops]   
        rprops_mean_intensity = np.asarray(rprops_mean_intensity) 
         
        rprops_label = [region['label'] for region in rprops]  
        rprops_label = np.asarray(rprops_label) 
             
        [ind]=np.where(rprops_mean_intensity>0) 
         
        im_out = np.isin(conncomp_t2, rprops_label[ind])    # test if the indice is in the matrix of indices 
        im_out =im_out*1    #  boolean to 0,1 conversion 
                     
        # Display 
        if display :  
            ylabel =kwargs.pop('ylabel','Frequency [Hz]') 
            xlabel =kwargs.pop('xlabel','Time [sec]')  
            title  =kwargs.pop('title','binary image => MASK') 
            cmap   =kwargs.pop('cmap','gray')  
            vmin=kwargs.pop('vmin',0)  
            vmax=kwargs.pop('vmax',1)  
            extent=kwargs.pop('extent',None)
                
            if extent is None : 
                xlabel = 'pseudotime [points]'
                ylabel = 'pseudofrequency [points]'
             
            _, fig = plot2d (im_out, 
                             extent = extent, 
                             title  = title,  
                             ylabel = ylabel, 
                             xlabel = xlabel,
                             vmin   = vmin, 
                             vmax   = vmax, 
                             cmap   = cmap, 
                             **kwargs)  
            
            # SAVE FIGURE 
            if savefig is not None :  
                dpi   =kwargs.pop('dpi',96) 
                format=kwargs.pop('format','png') 
                filename=kwargs.pop('filename','_spectro_binary')              
                filename = savefig+filename+'.'+format 
                if verbose :
                    print('\n''save figure : %s' %filename) 
                fig.savefig(filename, bbox_inches='tight', dpi=dpi, format=format, 
                            **kwargs)    
 
    return im_out 
 
#%% 
def _double_threshold_abs(im, bin_h=0.7, bin_l=0.2, 
                          verbose=False,display=False, savefig=None, **kwargs): 
    """ 
    Binarize an image based on a double relative threshold.  
    The values used for the thresholding are independent of the values in the 
     
    image => absolute threshold 
     
    Parameters 
    ---------- 
    im : 2d ndarray of scalars 
        Spectrogram (or image) 

    bin_h : scalar, optional, default is 0.7 
        Set the first threshold. Value higher than this value are set to 1,  
        the others are set to 0. They are the seeds for the second step 
         
    bin_l: scalar, optional, defautl is 0.2 
        Set the second threshold. Value higher than this value and connected 
        to the seeds or to other pixels connected to the seeds are set to 1,  
        the other remains 0 
        
    verbose : boolean, optional, default is False
        print messages
 
    display : boolean, optional, default is False 
        Display the signal if True 
         
    savefig : string, optional, default is None 
        Root filename (with full path) is required to save the figures. Postfix 
        is added to the root filename. 
         
    kwargs, optional. This parameter is used by plt.plot and savefig functions 
           
        - savefilename : str, optional, default :'_spectro_after_noise_subtraction.png' 
            Postfix of the figure filename 
         
        - figsize : tuple of integers, optional, default: (4,10) 
            width, height in inches.   
         
        - title : string, optional, default : 'Spectrogram' 
            title of the figure 
         
        - xlabel : string, optional, default : 'Time [s]' 
            label of the horizontal axis 
         
        - ylabel : string, optional, default : 'Amplitude [AU]' 
            label of the vertical axis 
         
        - cmap : string or Colormap object, optional, default is 'gray' 
            See https://matplotlib.org/examples/color/colormaps_reference.html 
            in order to get all the  existing colormaps 
            examples: 'hsv', 'hot', 'bone', 'tab20c', 'jet', 'seismic',  
            'viridis'... 
         
        - vmin, vmax : scalar, optional, default: None 
            `vmin` and `vmax` are used in conjunction with norm to normalize 
            luminance data.  Note if you pass a `norm` instance, your 
            settings for `vmin` and `vmax` will be ignored. 
         
        - extent : scalars (left, right, bottom, top), optional, default: None 
            The location, in data-coordinates, of the lower-left and 
            upper-right corners. If `None`, the image is positioned such that 
            the pixel centers fall on zero-based (row, column) indices. 
         
        - dpi : integer, optional, default is 96 
            Dot per inch.  
            For printed version, choose high dpi (i.e. dpi=300) => slow 
            For screen version, choose low dpi (i.e. dpi=96) => fast 
         
        - format : string, optional, default is 'png' 
            Format to save the figure 
             
        ... and more, see matplotlib    
 
    Returns 
    ------- 
    im_out: binary image  
     
    References 
    ---------- 
    .. [1] J. Canny. A computational approach to edge detection. IEEE Transactions on Pattern Analysis and Machine Intelligence. 1986; vol. 8, pp.679-698. `DOI: 10.1109/TPAMI.1986.4767851 <https://doi.org/10.1109/TPAMI.1986.4767851>`_
    """ 
     
    # binarisation  
    im_t1 = im > bin_h  # mask1 
    im_t2 = im > bin_l  # mask2 
    im_t3 = im * im_t1  # selected parts of the image 
     
    #find index of regions which meet the criteria 
    conncomp_t2 = measure.label(im_t2)    #Find connected components in binary image 
    rprops = measure.regionprops(conncomp_t2,im_t3)  
     
    rprops_mean_intensity = [region['mean_intensity'] for region in rprops]   
    rprops_mean_intensity = np.asarray(rprops_mean_intensity) 
     
    rprops_label = [region['label'] for region in rprops]  
    rprops_label = np.asarray(rprops_label) 
         
    [ind]=np.where(rprops_mean_intensity>0) 
     
    im_out = np.isin(conncomp_t2, rprops_label[ind])    # test if the indice is in the maxtrix of indices 
    im_out =im_out*1    #  boolean to 0,1 conversion 
    
    if verbose :
        print(72 * '_') 
        print('Double thresholding with  absolute values...') 
        print ('**********************************************************') 
        print ('  Number of rois %.2f | Rois cover %.2f%' % (len(rprops_label), 
                                                             sum(im_out)/(im_out.shape[1]*im_out.shape[0])*100)) 
        print ('**********************************************************') 
                 
    # Display 
    if display :  
        ylabel =kwargs.pop('ylabel','Frequency [Hz]') 
        xlabel =kwargs.pop('xlabel','Time [sec]')  
        title  =kwargs.pop('title','binary image => MASK') 
        cmap   =kwargs.pop('cmap','gray')  
        vmin=kwargs.pop('vmin',0)  
        vmax=kwargs.pop('vmax',1)  
        extent=kwargs.pop('extent',None)
        interpolation=kwargs.pop('interpolation','none')
                
        if extent is None : 
            xlabel = 'pseudotime [points]'
            ylabel = 'pseudofrequency [points]'
         
        _, fig = plot2d (im_out, 
                         extent = extent, 
                         title  = title,  
                         ylabel = ylabel, 
                         xlabel = xlabel,
                         vmin   = vmin, 
                         vmax   = vmax, 
                         cmap   = cmap, 
                         interpolation = interpolation,
                         **kwargs) 
        # SAVE FIGURE 
        if savefig is not None :  
            dpi   =kwargs.pop('dpi',96) 
            format=kwargs.pop('format','png') 
            filename=kwargs.pop('filename','_spectro_binary')              
            filename = savefig+filename+'.'+format 
            if verbose :
                print('\n''save figure : %s' %filename) 
            fig.savefig(filename, bbox_inches='tight', dpi=dpi, format=format, 
                        **kwargs)    
 
    return im_out 
 
#%%
# =============================================================================
# public functions
# =============================================================================    
def create_mask(im, mode_bin = 'relative', 
                verbose= False, display = False, savefig = None, **kwargs): 
    """ 
    Binarize an image based on a double threshold.  
     
    Parameters 
    ---------- 
    im : 2d ndarray of scalars 
        Spectrogram (or image) 
 
    mode_bin : string in {'relative', 'absolute'}, optional, default is 'relative' 
        if 'absolute' [1]_ , a double threshold with absolute value is performed 
        with two parameters (see kwargs section)
        if 'relative' [2]_, a relative double threshold is performed with two 
        parameters (see kwargs section)

    verbose : boolean, optional, default is False
        print messages
         
    display : boolean, optional, default is False 
        Display the signal if True 
         
    savefig : string, optional, default is None 
        Root filename (with full path) is required to save the figures. Postfix 
        is added to the root filename. 
         
    kwargs, optional. This parameter is used by the maad functions as well 
        as the plt.plot and savefig functions. 
        All the input arguments required or optional in the signature of the 
        functions above can be passed as kwargs : 
        
        if 'absolute' [1]_ 
        - bin_h : scalar, optional, default is 0.7 
        Set the first threshold. Value higher than this value are set to 1,  
        the others are set to 0. They are the seeds for the second step 
        - bin_l: scalar, optional, defautl is 0.2 
        Set the second threshold. Value higher than this value and connected 
        to the seeds or to other pixels connected to the seeds (6-connectivity)
        are set to 1, the other remains 0  
        
        if 'relative' [2]_ :
        - bin_std :  scalar, optional, default is 6 
        bin_std is needed to compute the threshold1. 
        This threshold is not an absolute value but depends on values that are 
        similar to 75th percentile (pseudo_mean) and a sort of std value of 
        the image.  
        threshold1 = "pseudo_mean" + "std" * bin_std    
        Value higher than threshold1 are set to 1, they are the seeds for  
        the second step. The others are set to 0.  
        - bin_per: scalar, optional, defautl is 0.5 
        Set how much the second threshold is lower than the first 
        threshold value. From 0 to 1. ex: 0.1 = 10 %. 
        threshold2 = threshold1 (1-bin_per)   
        Value higher than threshold2 and connected (6-connectivity) to the  
        seeds are set to 1, the other remains 0 
            
        ... and more, see matplotlib    
 
    Returns 
    ------- 
    im_bin: binary image 
    
    References
    ----------
    .. [1] J. Canny. A computational approach to edge detection. IEEE Transactions on Pattern Analysis and Machine Intelligence. 1986; vol. 8, pp.679-698. `DOI: 10.1109/TPAMI.1986.4767851 <https://doi.org/10.1109/TPAMI.1986.4767851>`_
    .. [2] from MATLAB: Threshold estimation (Oliveira et al, 2015) 
     
    Examples 
    -------- 

    >>> import maad
    
    Load audio recording and convert it into spectrogram
    
    >>> s, fs = maad.sound.load('../data/cold_forest_daylight.wav')
    >>> Sxx,tn,fn,ext = maad.sound.spectrogram (s, fs, fcrop=(0,10000))   
    
    Convert linear spectrogram into dB
    
    >>> Sxx_dB = maad.util.power2dB(Sxx) +96
    
    Smooth the spectrogram
    
    >>> Sxx_dB_blurred = maad.sound.smooth(Sxx_dB)
    
    Detection of the acoustic signature => creation of a mask
    
    >>> im_bin = maad.rois.create_mask(Sxx_dB_blurred, bin_std=1.5, bin_per=0.25, mode='relative') 
    
    Plot spectrograms
    
    >>> import matplotlib.pyplot as plt 
    >>> fig, (ax1, ax2) = plt.subplots(2, 1)
    >>> ax1, fig = maad.util.plot2d(Sxx_dB, ax=ax1, extent=ext, title='original', vmin=10, vmax=70)
    >>> ax2, fig = maad.util.plot2d(im_bin, ax=ax2, extent=ext, title='mask')
    >>> fig.set_size_inches(13,8)
    >>> fig.tight_layout() 
    
    """ 
        
    if mode_bin == 'relative': 
        bin_std=kwargs.pop('bin_std', 6)  
        bin_per=kwargs.pop('bin_per', 0.5)  
        im_bin = _double_threshold_rel(im, bin_std, bin_per, 
                                       verbose, display, savefig, **kwargs) 
         
    elif mode_bin == 'absolute': 
        bin_h=kwargs.pop('bin_h', 0.7)  
        bin_l=kwargs.pop('bin_l', 0.3)  
        im_bin = _double_threshold_abs(im, bin_h, bin_l,
                                       verbose, display, savefig, **kwargs)    
    
    return im_bin  

#%%
def select_rois(im_bin, min_roi=None ,max_roi=None, 
                verbose=False, display=False, savefig = None, **kwargs): 
    """ 
    Select regions of interest based on its dimensions.
    
    The input is a binary mask, and the output is an image with labelled pixels. 
 
    Parameters 
    ---------- 
    im : 2d ndarray of scalars 
        Spectrogram (or image) 
         
    min_roi, max_roi : scalars, optional, default : None 
        Define the minimum and the maximum area possible for an ROI. If None,  
        the minimum ROI area is 1 pixel and the maximum ROI area is the area of  
        the image 
        
    verbose : boolean, optional, default is False
        print messages
         
    display : boolean, optional, default is False 
        Display the signal if True 
         
    savefig : string, optional, default is None 
        Root filename (with full path) is required to save the figures. Postfix 
        is added to the root filename. 
         
    kwargs, optional. This parameter is used by plt.plot and savefig functions 
            
        - savefilename : str, optional, default :'_spectro_after_noise_subtraction.png' 
            Postfix of the figure filename 
         
        - figsize : tuple of integers, optional, default: (4,10) 
            width, height in inches.   
         
        - title : string, optional, default : 'Spectrogram' 
            title of the figure 
         
        - xlabel : string, optional, default : 'Time [s]' 
            label of the horizontal axis 
         
        - ylabel : string, optional, default : 'Amplitude [AU]' 
            label of the vertical axis 
         
        - cmap : string or Colormap object, optional, default is 'gray' 
            See https://matplotlib.org/examples/color/colormaps_reference.html 
            in order to get all the  existing colormaps 
            examples: 'hsv', 'hot', 'bone', 'tab20c', 'jet', 'seismic',  
            'viridis'... 
         
        - vmin, vmax : scalar, optional, default: None 
            `vmin` and `vmax` are used in conjunction with norm to normalize 
            luminance data.  Note if you pass a `norm` instance, your 
            settings for `vmin` and `vmax` will be ignored. 
         
        - extent : scalars (left, right, bottom, top), optional, default: None 
            The location, in data-coordinates, of the lower-left and 
            upper-right corners. If `None`, the image is positioned such that 
            the pixel centers fall on zero-based (row, column) indices. 
         
        - dpi : integer, optional, default is 96 
            Dot per inch.  
            For printed version, choose high dpi (i.e. dpi=300) => slow 
            For screen version, choose low dpi (i.e. dpi=96) => fast 
         
        - format : string, optional, default is 'png' 
            Format to save the figure 
             
        ... and more, see matplotlib    
 
    Returns 
    ------- 
    im_rois: 2d ndarray 
        image with labels as values 
             
    rois: pandas DataFrame 
        Regions of interest with future descriptors will be computed. 
        Array have column names: ``labelID``, ``label``, ``min_y``, ``min_x``,
        ``max_y``, ``max_x``,
        Use the function ``maad.util.format_features`` before using 
        centroid_features to format of the ``rois`` DataFrame 
        correctly.
        
    Examples 
    --------
    >>> import maad
    
    Load audio recording compute the spectrogram in dB.
    
    >>> s, fs = maad.sound.load('../data/cold_forest_daylight.wav')
    >>> Sxx,tn,fn,ext = maad.sound.spectrogram (s, fs, flims=(0,20000), display=True)           
    >>> Sxx_dB = maad.util.power2dB(Sxx) +96
    
    Smooth the spectrogram
    
    >>> Sxx_dB_blurred = maad.sound.smooth(Sxx_dB)
    
     Using image binarization, detect isolated region in the time-frequency domain with high density of energy, i.e. regions of interest (ROIs).
    
    >>> im_bin = maad.rois.create_mask(Sxx_dB_blurred, bin_std=1.5, bin_per=0.5, mode='relative')
    
    Select ROIs from the binary mask.
    
    >>> im_rois, df_rois = maad.rois.select_rois(im_bin, display=True)
    >>> print('Without subtracted the background noise, only {} ROIs have been segmented'.format(len(df_rois)))
    Without subtracted the background noise, only 13 ROIs have been segmented
    >>> df_rois.iloc[-1]
    labelID         13
    label      unknown
    min_y          162
    min_x         1105
    max_y          193
    max_x         1127
    Name: 12, dtype: object

    We detected the background noise as a ROI, and that multiple ROIs are mixed in a single region. To have better results, it is adviced to preprocess the spectrogram to remove the background noise before creating the mask.
    
    >>> Sxx_noNoise = maad.sound.median_equalizer(Sxx)
    >>> Sxx_noNoise_dB = maad.util.power2dB(Sxx_noNoise)     
    >>> Sxx_noNoise_dB_blurred = maad.sound.smooth(Sxx_noNoise_dB)        
    >>> im_bin2 = maad.rois.create_mask(Sxx_noNoise_dB_blurred, bin_std=6, bin_per=0.5, mode='relative') 
    >>> im_rois2, df_rois2 = maad.rois.select_rois(im_bin2, display=True)
    >>> print('{} ROIs have been segmented'.format(len(df_rois2)))
    274 ROIs have been segmented
    >>> df_rois2.iloc[-1]
    labelID        274
    label      unknown
    min_y          214
    min_x         4485
    max_y          230
    max_x         4494
    Name: 273, dtype: object
    """ 
 
    # test if max_roi and min_roi are defined 
    if max_roi is None:  
        # the maximum ROI is set to the aera of the image 
        max_roi=im_bin.shape[0]*im_bin.shape[1] 
         
    if min_roi is None: 
        # the min ROI area is set to 1 pixel 
        min_roi = 1 
    
    if verbose :
        print(72 * '_') 
        print('Automatic ROIs selection in progress...') 
        print ('**********************************************************') 
        print ('  Min ROI area %d pix² | Max ROI area %d pix²' % (min_roi, max_roi)) 
        print ('**********************************************************') 
 
    labels = measure.label(im_bin)    #Find connected components in binary image 
    rprops = measure.regionprops(labels) 
     
    rois_bbox = [] 
    rois_label = [] 
     
    for roi in rprops: 
         
        # select the rois  depending on their size 
        if (roi.area >= min_roi) & (roi.area <= max_roi): 
            # get the label 
            rois_label.append(roi.label) 
            # get rectangle coordonates            
            rois_bbox.append (roi.bbox)     
                 
    im_rois = np.isin(labels, rois_label)    # test if the indice is in the matrix of indices 
    im_rois = im_rois* labels 
     
    # create a list with labelID and labelName (None in this case) 
    rois_label = list(zip(rois_label,['unknown']*len(rois_label))) 
    
    # test if there is a roi
    if len(rois_label)>0 :
        # create a dataframe rois containing the coordonates and the label 
        rois = np.concatenate((np.asarray(rois_label), np.asarray(rois_bbox)), axis=1) 
        rois = pd.DataFrame(rois, columns = ['labelID', 'label', 'min_y','min_x','max_y', 'max_x']) 
        # force type to integer 
        rois = rois.astype({'label': str,'min_y':int,'min_x':int,'max_y':int, 'max_x':int}) 
        # compensate half-open interval of bbox from skimage 
        rois.max_y -= 1 
        rois.max_x -= 1 
        
    else :
        rois = []    
        rois = pd.DataFrame(rois, columns = ['labelID', 'label', 'min_y','min_x','max_y', 'max_x']) 
        rois = rois.astype({'label': str,'min_y':int,'min_x':int,'max_y':int, 'max_x':int}) 
     
    # Display 
    if display :  
        ylabel =kwargs.pop('ylabel','Frequency [Hz]') 
        xlabel =kwargs.pop('xlabel','Time [sec]')  
        title  =kwargs.pop('title','Selected ROIs')  
        extent=kwargs.pop('extent',None)
                
        if extent is None : 
            xlabel = 'pseudotime [points]'
            ylabel = 'pseudofrequency [points]'
         
        # randcmap = rand_cmap(len(rois_label)) 
        # cmap   =kwargs.pop('cmap',randcmap)  
        cmap   =kwargs.pop('cmap','tab20') 
         
        _, fig = plot2d (im_rois, 
                         interpolation = 'none',
                         extent = extent,
                         title  = title,  
                         ylabel = ylabel, 
                         xlabel = xlabel,
                         cmap   = cmap, 
                         **kwargs) 
        # SAVE FIGURE 
        if savefig is not None :  
            dpi   =kwargs.pop('dpi',96) 
            format=kwargs.pop('format','png')  
            filename=kwargs.pop('filename','_spectro_selectrois')                 
            filename = savefig+filename+'.'+format 
            fig.savefig(filename, bbox_inches='tight', dpi=dpi, format=format, 
                        **kwargs)  
            
    return im_rois, rois 
 
#%%
def rois_to_imblobs(im_zeros, rois): 
    """  
    Take a matrix full of zeros and add ones in delimited regions defined by rois.
 
    Parameters 
    ---------- 
    im_zeros : ndarray 
        matrix full of zeros with the size to the image where the rois come from.
     
    rois : DataFrame 
        rois must have the columns names:((min_y, min_x, max_y, max_x) which 
        correspond to the bounding box coordinates 
     
    Returns 
    ------- 
    im_blobs : ndarray 
        matrix with 1 corresponding to the rois and 0 elsewhere 
 
    Examples
    --------
    >>> from maad import rois, util
    >>> import pandas as pd
    >>> import numpy as np
    >>> im_zeros = np.zeros((100,300))
    >>> df_rois = pd.DataFrame({'min_y': [10, 40], 'min_x': [10, 200], 'max_y': [60, 80], 'max_x': [110, 250]})
    >>> im_blobs = rois.rois_to_imblobs(im_zeros, df_rois)
    >>> ax, fig = util.plot2d(im_blobs)
    """ 
    # Check format of the input data 
    if type(rois) is not pd.core.frame.DataFrame : 
        raise TypeError('Rois must be of type pandas DataFrame')   
         
    if not(('min_y' and 'min_x' and 'max_y' and 'max_x')  in rois)  : 
            raise TypeError('Array must be a Pandas DataFrame with column names:((min_y, min_x, max_y, max_x). Check example in documentation.')   
     
    # select the columns 
    rois_bbox = rois[['min_y', 'min_x', 'max_y', 'max_x']] 
    # roi to image blob 
    for min_y, min_x, max_y, max_x in rois_bbox.values: 
        im_zeros[int(min_y):int(max_y+1), int(min_x):int(max_x+1)] = 1 
     
    im_blobs = im_zeros.astype(int) 
     
    return im_blobs 

def spectrogram_local_max(
    Sxx, tn, fn, ext,
    min_distance,
    threshold_abs,
    display=False,
    **kwargs
):
    """
    Find peaks on spectrogram as coordinate list in time and frequency

    Parameters
    ----------
    Sxx : ndarray
        Spectrogram of audio signal.
    tn : 1d array
        Time vector of target audio, which results from the maad.sound.spectrogram function.
    fn : 1d array
        Frecuency vector of target audio, which results from the maad.sound.spectrogram function.
    ext : list of scalars [left, right, bottom, top]
        Extent keyword arguments controls the bounding box in data coordinates for the spectrogram of the target audio, which results from the maad.sound.spectrogram function.
    min_distance : int
        Minimum number of time-frequency coefficients (or pixels) separating peaks. This parameter controls how close peaks can be to each other. Peaks that are closer than min_distance will be merged into a single peak.
    threshold_abs : float
        Minimum amplitude threshold for peak detection. Must be above Sxx.min().
    display : bool, optional
        Option to display the resulting figure.

    Returns
    -------
    peak_time: numpy.array
        The temporal coordinates of local peaks (maxima) in a spectrogram. 
    peak_frequency: numpy.array
        The spectral coordinates of local peaks (maxima) in a spectrogram.

    Examples
    --------
    >>> from maad import sound, rois, util
    >>> s, fs = sound.load('../data/spinetail.wav')
    >>> Sxx, tn ,fn ,ext = sound.spectrogram (s, fs, nperseg=512, noverlap=256)
    >>> Sxx_db = util.power2dB(Sxx, db_range=80)
    >>> peak_time, peak_freq = rois.spectrogram_local_max(Sxx_db, tn, fn, ext, min_distance=1, threshold_abs=-40, display=True)
    """

    # Validate input
    if threshold_abs is not None:
        if threshold_abs < Sxx.min():
            raise ValueError(f'Value for minimum peak amplitude is below minimum value on spectrogram')

    # Find peaks in spectrogram
    peaks = peak_local_max(
        Sxx, min_distance=min_distance, threshold_abs=threshold_abs, **kwargs
    )

    if display == True:
        fig, ax = plt.subplots(nrows=1, figsize=(10, 5))
        plot_spectrogram(Sxx, ext, log_scale=False, db_range=80, ax=ax)
        ax.scatter(
            tn[peaks[:, 1]],
            fn[peaks[:, 0]],
            marker="o",
            facecolor="none",
            edgecolor="yellow",
        )

    return tn[peaks[:, 1]], fn[peaks[:, 0]] 

if __name__ == "__main__":
    import doctest
    doctest.testmod()



#!/usr/bin/env python
""" 
Collection of functions to load, read, write, play audio signal or 
its time-frequency representation
"""
#
# Authors:  Juan Sebastian ULLOA <lisofomia@gmail.com>
#           Sylvain HAUPERT <sylvain.haupert@mnhn.fr>
#
# License: New BSD License

# =============================================================================
# Load the modules
# =============================================================================
# Import external modules
import numpy as np
import warnings
import io
from urllib.request import urlopen
from scipy.io import wavfile 
from skimage.io import imread 
from scipy.io.wavfile import write as write_wav

# Import internal modules
from maad.util import plot1d, plot2d, linear_scale

#%%
# =============================================================================
# public functions
# =============================================================================
def load(filename, channel='left', detrend=True, verbose=False,
         display=False, savefig=None, **kwargs): 
    """
    Load an audio file (stereo or mono). 
    
    Currently, this function con only load WAVE files.
    
    Parameters
    ----------
    filename : string 
        Name or path of the audio file
    channel : {'left', right'}, optional, default: left
        In case of stereo sound select the channel that is kept 
    detrend : boolean, optional, default is True
        Subtract the DC value.
    verbose : boolean, optional, default is False
        Print messages into the console or terminal if verbose is True
    display : boolean, optional, default is False
        Display the signal if True
    savefig : string, optional, default is None
        Root filename (with full path) is required to save the figures. Postfix
        is added to the root filename.
    kwargs, optional. This parameter is used by plt.plot and savefig functions    
        - savefilename : str, optional, default :'_audiogram.png'
            Postfix of the figure filename
        - figsize : tuple of integers, optional, default: (4,10)
            width, height in inches.  
        - title : string, optional, default : 'Spectrogram'
            title of the figure
        - xlabel : string, optional, default : 'Time [s]'
            label of the horizontal axis
        - ylabel : string, optional, default : 'Amplitude [AU]'
            label of the vertical axis
        - cmap : string or Colormap object, optional, default is 'gray'
            See https://matplotlib.org/examples/color/colormaps_reference.html
            in order to get all the  existing colormaps
            examples: 'hsv', 'hot', 'bone', 'tab20c', 'jet', 'seismic', 
            'viridis'...
        - vmin, vmax : scalar, optional, default: None
            `vmin` and `vmax` are used in conjunction with norm to normalize
            luminance data.  Note if you pass a `norm` instance, your
            settings for `vmin` and `vmax` will be ignored.
            extent : list of scalars [left, right, bottom, top], optional, default: None
            The location, in data-coordinates, of the lower-left and
            upper-right corners. If `None`, the image is positioned such that
            the pixel centers fall on zero-based (row, column) indices.
        - dpi : integer, optional, default is 96
            Dot per inch. 
            For printed version, choose high dpi (i.e. dpi=300) => slow
            For screen version, choose low dpi (i.e. dpi=96) => fast
        - format : string, optional, default is 'png'
            Format to save the figure

        ... and more, see matplotlib     
        
    Returns
    -------    
    s_out : 1d ndarray of integer
        Vector containing the audiogram 
    fs : int 
        The sampling frequency in Hz of the audiogram    
        
    Examples
    --------
    >>> import maad
    >>> s, fs = maad.sound.load("../data/tropical_forest_morning.wav", channel='left')
    >>> print("The sampling frequency of the audio file is {} Hz".format(fs))
    The sampling frequency of the audio file is 48000 Hz
    >>> import numpy as np
    >>> tn = np.arange(0,len(s))/fs
    >>> import matplotlib.pyplot as plt
    >>> fig, (ax0, ax1) = plt.subplots(2,1, sharex=True, squeeze=True)
    >>> ax0, _ = maad.util.plot1d(tn,s,ax=ax0, figtitle='ground level')
    >>> ax0.set_ylim((-0.075,0.075))
    (-0.075, 0.075)
    >>> s, fs = maad.sound.load("../data/tropical_forest_morning.wav", channel='right')
    >>> ax1, _ = maad.util.plot1d(tn,s,ax=ax1, figtitle='canopy level')
    >>> ax1.set_ylim((-0.075,0.075))
    (-0.075, 0.075)
    >>> fig.tight_layout()
    """
    if verbose :
        print(72 * '_' )
        print("loading %s..." %filename)   
    
    # Read the .wav file and return the sampling frequency fs (Hz) 
    # and the audiogram s as a 1D array of integer
    # Execute the function wavfile.read with warnings suppressed in order to avoid the 
    # common warning "WavFileWarning: Chunk (non-data) not understood, skipping it"
    with warnings.catch_warnings():
        warnings.filterwarnings("ignore")  # Ignore all warnings
        fs, s = wavfile.read(filename)
        if verbose :print("Sampling frequency: %dHz" % fs)
    
    # Normalize the signal between -1 to 1 depending on the type (number of bits)
    if s.dtype == np.int32:
        bit = 32
        s = s/2**(bit-1)
    elif s.dtype == np.int16:
        bit = 16
        s = s/2**(bit-1)
    elif s.dtype == np.uint8:
        bit = 8
        s = s/2**(bit) # as it's unsigned
    
    # test if stereo signal. if YES => keep only the ch_select
    if s.ndim==2 :
        if channel == 'left' :
            if verbose :print("Select left channel")
            s_out = s[:,0] 
        else:
            if verbose :print("Select right channel")
            s_out = s[:,1] 
    else:
        s_out = s;
        
    # Detrend the signal by removing the DC offset
    if detrend: s_out = s_out - np.mean(s_out)
    
    # Time vector
    tn = np.arange(s_out.size)/fs 
    
    # DISPLAY
    if display : 
        figtitle=kwargs.pop('figtitle', 'Orignal sound')        
        _, fig = plot1d(tn, s_out, figtitle=figtitle, **kwargs)
        # SAVE FIGURE
        if savefig is not None : 
            dpi=kwargs.pop('dpi', 96) 
            bbox_inches=kwargs.pop('bbox_inches', 'tight') 
            format=kwargs.pop('format','png')
            savefilename=kwargs.pop('savefilename', '_audiogram')  
            filename = savefig+savefilename+'.'+format
            if verbose :print('\n''save figure : %s' %filename)
            fig.savefig(fname=filename, dpi=dpi, bbox_inches=bbox_inches,
                        format=format, **kwargs)  
            
    return s_out, fs

#%%
def load_spectrogram(filename, fs, duration, flims = None, flipud = True,
                verbose=False, display=False, **kwargs): 
    """ 
    Load an image from a file or an URL 

    Parameters 
    ----------   
    filename : string 
        Image file name, e.g. ``test.jpg`` or URL. 

    fs : scalar 
        Sampling frequency of the audiogram (in Hz) 

    duration : scalar 
        Duration of the audiogram (in s) 
        
    flims : list of 2 scalars [min, max], optional, default is None
        flims corresponds to the min and max boundary frequency values
        
    flipud : boolean, optional, default is True 
        Vertical flip of the matrix (image) 
        
    verbose : boolean, optional, default is False
        if True, print message in terminal

    display : boolean, optional, default is False 
        if True, display the image 

    kwargs, optional. This parameter is used by plt.plot  
        - figsize : tuple of integers, optional, default: (4,10) 
            width, height in inches.   

        - title : string, optional, default : 'Spectrogram' 
            title of the figure 

        - xlabel : string, optional, default : 'Time [s]' 
            label of the horizontal axis 

        - ylabel : string, optional, default : 'Amplitude [AU]' 
            label of the vertical axis 

        - cmap : string or Colormap object, optional, default is 'gray' 
            See https://matplotlib.org/examples/color/colormaps_reference.html 
            in order to get all the  existing colormaps 
            examples: 'hsv', 'hot', 'bone', 'tab20c', 'jet', 'seismic',  
            'viridis'... 

        - vmin, vmax : scalar, optional, default: None 
            `vmin` and `vmax` are used in conjunction with norm to normalize 
            luminance data.  Note if you pass a `norm` instance, your 
            settings for `vmin` and `vmax` will be ignored. 

        - extent : scalars (left, right, bottom, top), optional, default: None 
            The location, in data-coordinates, of the lower-left and 
            upper-right corners. If `None`, the image is positioned such that 
            the pixel centers fall on zero-based (row, column) indices. 

        - dpi : integer, optional, default is 96 
            Dot per inch.  
            For printed version, choose high dpi (i.e. dpi=300) => slow 
            For screen version, choose low dpi (i.e. dpi=96) => fast 

        - format : string, optional, default is 'png' 
            Format to save the figure  

        ... and more, see matplotlib  

    Returns 
    ------- 
    Sxx : ndarray 
        The different color bands/channels are stored in the 
        third dimension, such that a gray-image is MxN, an 
        RGB-image MxNx3 and an RGBA-image MxNx4.    
    tn : 1d ndarray of floats
        time vector (horizontal x-axis)    
    fn : 1d ndarray of floats
        Frequency vector (vertical y-axis)    
    extent : list of scalars [left, right, bottom, top]
        The location, in data-coordinates, of the lower-left and
        upper-right corners. 
        
    Examples
    --------
    >>> import maad
    >>> xenocanto_link = 'https://www.xeno-canto.org/sounds/uploaded/DTKJSKMKZD/ffts/XC445081-med.png'
    >>> Sxx, tn, fn, ext = maad.sound.load_spectrogram(filename=xenocanto_link, \
                                                        fs=44100,               \
                                                        flims=[0,15000],        \
                                                        duration = 10,          \
                                                        )
    >>> print("The time resolution of the spectrogram is {} s and the frequency resolution is {} Hz".format(tn[1]-tn[0], fn[1]-fn[0]))
    The time resolution of the spectrogram is 0.020876826722338204 s and the frequency resolution is 94.33962264150944 Hz
    >>> import matplotlib.pyplot as plt
    >>> ax, fig = maad.util.plot2d(Sxx,extent=ext)
    
    """ 
    
    if verbose :
        print(72 * '_' ) 
        print("loading %s..." %filename)  
    
    # Load image 
    Sxx = imread(filename, as_gray=True) 
    
    # if 3D, convert into 2D 
    if len(Sxx.shape) == 3: 
        Sxx = Sxx[:,:,0] 
    
    # Rescale the image between 0 to 1 
    Sxx = linear_scale(Sxx, minval= 0.0, maxval=1.0, axis=None) 
    
    # Get the resolution 
    if flims is None :
        df = fs/(Sxx.shape[0]-1) 
    else:
        df = (flims[1]-flims[0]) / (Sxx.shape[0]-1)
    dt = duration/(Sxx.shape[1]-1) 
    
    # create the vectors
    if flims is None :
        fn = np.arange(0,fs/2,df) 
    else:
        fn = np.arange(flims[0],flims[1],df)  
    tn = np.arange(0,Sxx.shape[1],1) * dt
    
    # Extent 
    extent = [tn[0], tn[-1], fn[0], fn[-1]] 
    
    # flip the image vertically 
    if flipud: Sxx = np.flip(Sxx, 0) 
    
    # Display 
    if display :  
        ylabel =kwargs.pop('ylabel','Frequency [Hz]') 
        xlabel =kwargs.pop('xlabel','Time [sec]')  
        title  =kwargs.pop('title','loaded spectrogram') 
        cmap   =kwargs.pop('cmap','gray')  
        figsize=kwargs.pop('figsize',(4, 0.33*(extent[1]-extent[0])))  
        vmin=kwargs.pop('vmin',Sxx.min())  
        vmax=kwargs.pop('vmax',Sxx.max())  
        
        _, fig = plot2d (Sxx, extent=extent, figsize=figsize,title=title,  
                        ylabel = ylabel, xlabel = xlabel,vmin=vmin, vmax=vmax, 
                        cmap=cmap, **kwargs) 
    
    return Sxx, tn, fn, extent 
#%%
def write(filename, fs, data, bit_depth=None):
    """
    Write a NumPy array as a WAV file with the Scipy method. [1]_ 

    Parameters
    ----------
    filename : string or open file handle
        Name of output wav file.
    fs : int
        Sample rate (samples/sec).
    data : ndarray
        Mono or stereo signal as NumPy array.
    bit_depth: int
        Specifies the bit depth format of the audio recording.
        Should be one of 8, 16 or 32. If None, bit depth will be determined from
        the Numpy data type. See section Common data types.
        
    See Also
    --------
    scipy.io.wavfile.write

    Notes
    -----
    The data-type determines the bits-per-sample and PCM/float.

    Common data types: [2]_

    =====================  ===========  ===========  =============
         WAV format            Min          Max       NumPy dtype
    =====================  ===========  ===========  =============
    32-bit floating-point  -1.0         +1.0         float32
    32-bit PCM             -2147483648  +2147483647  int32
    16-bit PCM             -32768       +32767       int16
    8-bit PCM              0            255          uint8
    =====================  ===========  ===========  =============

    References
    ----------
    .. [1] The SciPy community, "scipy.io.wavfile.write", v1.6.0.
       https://docs.scipy.org/doc/scipy/reference/generated/scipy.io.wavfile.write.html
       
    .. [2] IBM Corporation and Microsoft Corporation, "Multimedia Programming
       Interface and Data Specifications 1.0", section "Data Format of the
       Samples", August 1991
       http://www.tactilemedia.com/info/MCI_Control_Info.html

    Examples
    --------
    >>> from maad import sound
    
    Synthesize a 440Hz sine wave at 44100 Hz and write it to disk.
    
    >>> import numpy as np
    >>> fs = 44100; T = 2.0
    >>> t = np.linspace(0, T, int(T*fs))
    >>> data = np.sin(2. * np.pi * 440. *t)
    >>> sound.write('example.wav', fs, data, bit_depth=16)
    
    Open an audio file, filter a frequency band and write to disk specifying the bit depth.

    >>> s, fs = sound.load('../data/spinetail.wav')
    >>> s_filt = sound.sinc(s, (3000, 10000), fs)
    >>> sound.write('spinetail_filtered.wav', fs, s_filt, bit_depth=16)
    """

    if (data.dtype == 'float64') | (data.dtype == 'float32'):
        # Check that the array has values between -1 and 1
        if (data.min() < -1) | (data.max() > 1):
            warnings.warn('Values should be between [-1, 1]. Clipping signal.')
            data[data<-1] = -1
            data[data>1] = 1
        
        # Convert to desired bit depth
        if bit_depth == 8:
            data = data + 1 # change range to postive [0,2]
            data = (data * 127).astype(np.uint8)
            
        elif bit_depth == 16:
            data = (data * 32767).astype(np.int16)
            
        elif bit_depth == 32:
            data = (data * 2147483647).astype(np.int32)
        
        else:
            warnings.warn('Values for bit depth should be 8, 16 or 32. Argument ignored.')
            pass
        
    if data.ndim > 1:
        data = data.T
    
    write_wav(filename, fs, np.asfortranarray(data))
    
#%%
def load_url(url):
    """
    Download audio file from the web and load it as a variable. The audio file must be
    a Waveform Audio Format (WAV) file.
    
    Parameters
    ----------
    url : str
        Audio code name or URL address where the file is located.
    
    Returns
    -------
    s : 1d ndarray
        Array with the signal amplitude
    fs : int
        Sampling frequency of the signal, in Hertz (Hz)
    
    Examples
    --------
    
    Load an audio example using its code name.
    
    >>> from maad import sound
    >>> s, fs = sound.load_url('spinetail')
    >>> print('The samping rate of the audio file is {} Hz'.format(fs))
    The samping rate of the audio file is 44100 Hz

    Load an audio example using the full web address.
    
    >>> s, fs = sound.load_url('https://github.com/scikit-maad/scikit-maad/raw/production/data/spinetail.wav')
    >>> print('The samping rate of the audio file is {} Hz'.format(fs))
    The samping rate of the audio file is 44100 Hz
    """

    # set dictionary for examples from the audio dataset
    code_name_url = {
        'spinetail': 'https://github.com/scikit-maad/scikit-maad/raw/production/data/spinetail.wav',
        'cold_forest_daylight': 'https://github.com/scikit-maad/scikit-maad/raw/production/data/cold_forest_daylight.wav',
        'cold_forest_night': 'https://github.com/scikit-maad/scikit-maad/raw/production/data/cold_forest_night.wav',
        'rock_savanna': 'https://github.com/scikit-maad/scikit-maad/raw/production/data/guyana_tropical_forest.wav',
        'tropical_forest_morning': 'https://github.com/scikit-maad/scikit-maad/raw/production/data/tropical_forest_morning.wav'
        }
    
    # Check if url is in the dictionary keys    
    if url in code_name_url.keys():
        url_path = code_name_url[url]
    # load as web address
    else:
        url_path = url
    
    s, fs = load(io.BytesIO(urlopen(url_path).read()))
    return s, fs


if __name__ == "__main__":
    import doctest
    doctest.testmod()


#!/usr/bin/env python
"""
Collection of functions to extract features from audio signals
License: New BSD License
"""

# =============================================================================
# Load the modules
# =============================================================================
# Import external modules
import numpy as np
import pandas as pd
from scipy import interpolate
from scipy.optimize import root

# Import internal modules
from maad.util import moments
from maad.sound import envelope, trim
from maad import sound

# define internal functions
def _quantile_is_valid(q):
    """ Check if quantile is valid
        function from older version of numpy than 2.0.0
    """
    # avoid expensive reductions, relevant for arrays with < O(1000) elements
    if q.ndim == 1 and q.size < 10:
        for i in range(q.size):
            if not (0.0 <= q[i] <= 1.0):
                return False
    else:
        if not (np.all(0 <= q) and np.all(q <= 1)):
            return False
    return True


#%%
# =============================================================================
# public functions
# =============================================================================
#%%
def temporal_moments(s, fs=None, roi=None):
    """
    Computes the first 4th moments of an audio signal, mean, variance, skewness, kurtosis.

    Parameters
    ----------
    s : 1D array
        Audio to process
    fs : float, optional
        Sampling frequency of audio signal
        The default is None
    roi : pandas.Series, optional
        Region of interest where peak frequency will be computed.
        Series must have a valid input format with index: min_t, min_f, max_t, max_f.
        The default is None.        

    Returns
    -------
    mean : float
        mean of the audio
    var : float
        variance  of the audio
    skew : float
        skewness of the audio
    kurt : float
        kurtosis of the audio

    Examples
    --------
    >>> from maad import sound, features
    >>> s, fs = sound.load('../data/spinetail.wav')
    >>> sm, sv, ss, sk = features.temporal_moments (s, fs)
    >>> print('mean: %2.2f / var: %2.5f / skewness: %2.4f / kurtosis: %2.2f' % (sm, sv, ss, sk))
    mean: -0.00 / var: 0.00117 / skewness: -0.0065 / kurtosis: 24.71

    """
    # force s to be ndarray
    s = np.asarray(s)

    if (roi is not None):
        if (fs is not None) :
            s = trim(s, fs, min_t=roi.min_t, max_t=roi.max_t)
        else : 
            raise ValueError("If 'roi' is not None, 'fs' must be defined")     
        
    return moments(s)

#%%
def zero_crossing_rate(s, fs, roi=None):
    """
    Compute the zero crossing rate feature of an audio signal.

    The zero-crossing rate is the rate of sign-changes along a signal, i.e.,
    the rate at which the signal changes from positive to zero to negative or
    from negative to zero to positive. This feature has been used widely
    in speech recognition and music information retrieval,
    being a key feature to classify percussive sounds [1]_.

    Parameters
    ----------
    s : 1D array
        Audio to process (wav)
    fs : float
        Sampling frequency of audio signal
    roi : pandas.Series, optional
        Region of interest where peak frequency will be computed.
        Series must have a valid input format with index: min_t, min_f, max_t, max_f.
        The default is None.        

    Returns
    -------
    zcr : float
        number of zero crossing /s

    References
    ----------
    .. [1] https://en.wikipedia.org/wiki/Zero-crossing_rate

    Examples
    --------
    >>> from maad import sound, features
    >>> s, fs = sound.load('../data/spinetail.wav')
    >>> features.zero_crossing_rate(s,fs)
    10500.397192384766

    """

    if (roi is not None):
        s = trim(s, fs, min_t=roi.min_t, max_t=roi.max_t)

    zero_crosses = np.nonzero(np.diff(s > 0))[0]
    duration = len(s) / fs
    zcr = 1/duration * len(zero_crosses)

    return zcr

#%%
def temporal_quantile(s, fs, q=[0.05, 0.25, 0.5, 0.75, 0.95], nperseg=1024, roi=None, mode="spectrum",
                      env_mode="fast", as_pandas=False, amp=False, **kwargs):
    """
    Compute the q-th temporal quantile of the waveform or spectrum. If a
    region of interest with time and spectral limits is provided, the q-th
    temporal quantile is computed on the selection.

    Parameters
    ----------
    s : 1D array
        Input audio signal
    fs : float
        Sampling frequency of audio signal
    q : array or float, optional
        Quantile or sequence of quantiles to compute, which must be between 0 and 1
        inclusive.
        The defaul is [0.05, 0.25, 0.5, 0.75, 0.95].
    nperseg : int, optional
        Length of segment to compute the FFT when mode is spectrum. The default is 1024.
        Size of each frame to compute the envelope. The largest, the highest is
        the approximation. The default is 5000.
    roi : pandas.Series, optional
        Region of interest where peak frequency will be computed.
        Series must have a valid input format with index: min_t, min_f, max_t, max_f.
        The default is None.
    mode : str, optional, default is 'spectrum'
        - 'spectrum' : The quantile is calculated in the espectrum.
        - 'envelope' : The quantile is calculated in the sound wave.
    env_mode : str, optional, default is `fast`
        - `fast` : The sound is first divided into frames (2d) using the
            function wave2timeframes(s), then the max of each frame gives a
            good approximation of the envelope.
        - `Hilbert` : estimation of the envelope from the Hilbert transform.
            The method is slow
    as_pandas: bool
        Return data as a pandas.Series. This is usefull when computing multiple features
        over a signal. Default is False.
    amp: bool, default is False
        Return the quantiles with its amplitude.

    Returns
    -------
    quantiles: pandas Series/DataFrame or Numpy array
        Temporal quantiles of waveform and its amplitude (optional).

    Examples
    --------
    >>> from maad import features, sound
    >>> s, fs = sound.load('../data/spinetail.wav')

    Compute the q-th temporal quantile in the spectrum

    >>> qt = features.temporal_quantile(s, fs, [0.05, 0.25, 0.5, 0.75, 0.95], as_pandas=True)
    >>> print(qt)
    0.05     1.219048
    0.25     5.712109
    0.50    11.818957
    0.75    16.555828
    0.95    17.751655
    dtype: float64

    Compute the q-th temporal quantile in the waveform, using the envelope

    >>> qt = features.temporal_quantile(s, fs, [0.05, 0.25, 0.5, 0.75, 0.95], mode="envelope", as_pandas=True)
    >>> print(qt)
    0.05     1.208300
    0.25     5.716188
    0.50    11.804161
    0.75    15.731135
    0.95    17.752714
    dtype: float64

    """
    q = np.asanyarray(q)
    if not _quantile_is_valid(q):
        raise ValueError("Percentiles must be in the range [0, 1]")

    # compute quantiles in the time amp
    if mode=="envelope":
        if roi is None:
            min_t = 0
        elif fs is not None:
            s = sound.trim(s, fs, min_t=roi.min_t, max_t=roi.max_t)
            min_t = roi.min_t
        else :
            raise ValueError("If 'roi' is not None, 'fs' must be defined")

        env = sound.envelope(s**2, env_mode, nperseg)
        t = min_t+np.arange(0,len(env),1)*len(s)/fs/len(env)
        energy = pd.Series(env, index=t)

        # Compute temporal quantile
        norm_cumsum = energy.cumsum()/energy.sum()
        spec_quantile = []
        for quantile in q:
            spec_quantile.append(energy.index[np.where(norm_cumsum>=quantile)[0][0]])

        if amp:
            if as_pandas:
                out = pd.DataFrame({"time":spec_quantile, "amp":energy[spec_quantile].values}, index=q)
            else:
                out = np.transpose(np.array([q, spec_quantile, energy[spec_quantile]]))
        else:
            if as_pandas:
                out = pd.Series(spec_quantile, index=q)
            else:
                out = np.array(spec_quantile)

        return out

    elif mode=="spectrum":
        if roi is None:
            Sxx,tn,_,_ = sound.spectrogram(s, fs, nperseg=nperseg, **kwargs)
        else:
            Sxx,tn,_,_ = sound.spectrogram(s, fs, nperseg=nperseg,
                                        tlims=[roi.min_t, roi.max_t],
                                        flims=[roi.min_f, roi.max_f],
                                        **kwargs)
        Sxx = pd.Series(np.average(Sxx,axis=0), index=tn)

        # Compute spectral q
        norm_cumsum = Sxx.cumsum()/Sxx.sum()
        spec_quantile = []
        for quantile in q:
            spec_quantile.append(Sxx.index[np.where(norm_cumsum>=quantile)[0][0]])

        if amp:
            if as_pandas:
                out = pd.DataFrame({"time":spec_quantile, "amp":Sxx[spec_quantile].values}, index=q)
            else:
                out = np.transpose(np.array([q, spec_quantile, Sxx[spec_quantile]]))
        else:
            if as_pandas:
                out = pd.Series(spec_quantile, index=q)
            else:
                out = np.array(spec_quantile)

        return out
    else:
        raise Exception("Invalid mode. Mode should be 'spectrum' or 'envelope'")
    
#%%
def temporal_duration(s, fs, nperseg=1024, roi=None, mode="spectrum",
                      env_mode="fast", as_pandas=False, **kwargs):
    """
    Compute the temporal duration of the waveform. If a region of interest with time
    and spectral limits is provided, the temporal duration is computed on the selection.

    Parameters
    ----------
    s : 1D array
        Input audio signal
    fs : float
        Sampling frequency of audio signal
    nperseg : int, optional
        Length of segment to compute the FFT. The default is 1024.
    roi : pandas.Series, optional
        Region of interest where temporal duration will be computed.
        Series must have a valid input format with index: min_t, min_f, max_t, max_f.
        The default is None.
    mode : str, optional, default is 'spectrum'
        - 'spectrum' : The quantile is calculated using the spectrum.
        - 'amplitude' : The quantile is calculated using the enveloppe sound wave.
    env_mode : str, optional, default is `fast`
        - `fast` : The sound is first divided into frames (2d) using the
            function wave2timeframes(s), then the max of each frame gives a
            good approximation of the envelope.
        - `Hilbert` : estimation of the envelope from the Hilbert transform.
            The method is slow
    as_pandas_series: bool
        Return data as a pandas.Series. This is usefull when computing multiple features
        over a signal. Default is False.

    Returns
    -------
    duration: pandas Series/DataFrame or Numpy array
        Temporal duration of signal using energy quantiles.

    Examples
    --------
    >>> from maad import features, sound
    >>> s, fs = sound.load('../data/spinetail.wav')

    Compute the temporal duration of the time energy

    >>> duration, duration_90 = features.temporal_duration(s, fs)
    >>> print("Duration 50%: {:.4f} / Duration 90%: {:.4f}".format(duration, duration_90))
    Duration 50%: 10.8437 / Duration 90%: 16.5326

    """
    # Compute temporal quantile
    q = temporal_quantile(s, fs, [0.05, 0.25, 0.75, 0.95], nperseg, roi, mode, 
                          env_mode, as_pandas, **kwargs)

    # Compute temporal duration
    if as_pandas:
        out = pd.Series([np.abs(q.iloc[2]-q.iloc[1]), np.abs(q.iloc[3]-q.iloc[0])],
                        index=["duration_50", "duration_90"])
    else:
        out = np.array([np.abs(q[2]-q[1]), np.abs(q[3]-q[0])])

    return out

#%%
def all_temporal_features(s, fs, nperseg=1024, roi=None, display=False, **kwargs):
    """
    Compute all the temporal features for a signal.

    Parameters
    ----------
    s : 1D array
        Input audio signal
    fs : float
        Sampling frequency of audio signal
    nperseg : int, optional
        Length of segment to compute the FFT. The default is 1024.
    roi : pandas.Series, optional
        Region of interest where temporal features will be computed.
        Series must have a valid input format with index: min_t, min_f, max_t, max_f.
        The default is None.
    kwargs : additional keyword arguments
        If `window='hann'`, additional keyword arguments to pass to
        `sound.spectrum`.

    Returns
    -------
    temporal_features : pandas DataFrame
        DataFrame with all temporal features computed in the spectrum
    
    Examples
    --------
    >>> from maad import features, sound
    >>> s, fs = sound.load('../data/spinetail.wav')

    Compute all the temporal features

    >>> temporal_features = features.all_temporal_features(s,fs)
    >>> print(temporal_features.iloc[0])
    sm            -2.043264e-19
    sv             1.167074e-03
    ss            -6.547980e-03
    sk             2.471161e+01
    Time 5%        1.219048e+00
    Time 25%       5.712109e+00
    Time 50%       1.181896e+01
    Time 75%       1.655583e+01
    Time 95%       1.775166e+01
    zcr            1.050040e+04
    duration_50    1.001495e+01
    duration_90    1.654441e+01
    Name: 0, dtype: float64
    """

    tm  = temporal_moments(s, fs, roi)
    zcr = zero_crossing_rate(s, fs, roi)
    qt = temporal_quantile(s, fs, [0.05, 0.25, 0.5, 0.75, 0.95], nperseg, roi, mode="spectrum", **kwargs)
    duration_50, duration_90 = temporal_duration(s, fs, nperseg, roi, mode="envelope")

    temporal_features = pd.DataFrame({"sm":tm[0], "sv":tm[1], "ss":tm[2], "sk":tm[3],
                                    "Time 5%":qt[0], "Time 25%":qt[1], "Time 50%":qt[2], 
                                    "Time 75%":qt[3], "Time 95%":qt[4], "zcr":zcr,
                                    "duration_50":duration_50, "duration_90":duration_90}, index=[0])

    if display: print(temporal_features)

    return temporal_features

if __name__ == "__main__":
    import doctest
    doctest.testmod()